﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace San9.Resource.IO {
  public class BitReader {
    private readonly byte[] source;
    private readonly int offset;
    private readonly int length;
    private int index = 0;

    public BitReader(IEnumerable<byte> source) {
      this.source = source.ToArray();
      this.offset = 0;
      this.length = this.source.Length * 8;
    }

    private BitReader(byte[] source, int offset, int length) {
      this.source = source;
      this.offset = offset;
      this.length = length;
    }

    public bool this[int i] {
      get {
        if (i < 0 || length <= i) throw new ArgumentOutOfRangeException();
        return (source[(offset + i) / 8] & (0x80 >> ((offset + i) % 8))) != 0;
      }
    }

    public int Index {
      get { return index; }
    }

    public int Rest {
      get { return length - index; }
    }

    public bool AtEnd {
      get { return Rest == 0; }
    }

    public bool Current {
      get {
        if (AtEnd) throw new InvalidOperationException();
        return this[index];
      }
    }

    public bool Next() {
      if (AtEnd) throw new InvalidOperationException();
      index++;
      return !AtEnd;
    }

    public void Reset() {
      index = 0;
    }

    public bool[] ToArray() {
      bool[] array = new bool[Rest];
      for (int i = 0; i < Rest; i++)
        array[i] = this[index + i];
      return array;
    }

    public BitReader Take(int n) {
      if (n < 0 || Rest < n) throw new ArgumentOutOfRangeException();
      return new BitReader(source, offset + index, n);
    }

    public BitReader TakeWhile(Func<bool, bool> f) {
      int n = 0;
      while (n < Rest && f(this[index + n])) n++;
      return Take(n);
    }

    public BitReader TakeUntil(Func<bool, bool> f) {
      int n = 0;
      while (n < Rest && !f(this[index + n])) n++;
      if (n < Rest) n++;
      return Take(n);
    }

    public BitReader Drop(int n) {
      if (n < 0 || Rest < n) throw new ArgumentOutOfRangeException();
      return new BitReader(source, offset + index + n, Rest - n);
    }

    public BitReader DropWhile(Func<bool, bool> f) {
      int n = 0;
      while (n < Rest && f(this[index + n])) n++;
      return Drop(n);
    }

    public BitReader DropUntil(Func<bool, bool> f) {
      int n = 0;
      while (n < Rest && !f(this[index + n])) n++;
      if (n < Rest) n++;
      return Drop(n);
    }

    public void Skip(int n) {
      if (n > Rest) throw new ArgumentOutOfRangeException();
      index += n;
    }
  }
}
