﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace San9.Resource.IO {
  public class ByteReader {
    private readonly byte[] source;
    private readonly int offset;
    private readonly int length;
    private int index = 0;

    public ByteReader(IEnumerable<byte> source) {
      this.source = source.ToArray();
      this.offset = 0;
      this.length = this.source.Length;
    }

    private ByteReader(byte[] source, int offset, int length) {
      this.source = source;
      this.offset = offset;
      this.length = length;
    }

    public byte this[int i] {
      get {
        if (i < 0 || length <= i) throw new ArgumentOutOfRangeException();
        return source[offset + i];
      }
    }

    public int Index {
      get { return index; }
    }

    public int Rest {
      get { return length - index; }
    }

    public bool AtEnd {
      get { return Rest == 0; }
    }

    public byte Current {
      get {
        if (AtEnd) throw new InvalidOperationException();
        return this[index];
      }
    }

    public bool Next() {
      if (AtEnd) throw new InvalidOperationException();
      index++;
      return !AtEnd;
    }

    public void Reset() {
      index = 0;
    }

    public byte[] ToArray() {
      byte[] array = new byte[Rest];
      for (int i = 0; i < Rest; i++)
        array[i] = this[Index + i];
      return array;
    }

    public ByteReader Take(int n) {
      if (n < 0 || Rest < n) throw new ArgumentOutOfRangeException();
      return new ByteReader(source, offset + index, n);
    }

    public ByteReader TakeWhile(Func<byte, bool> f) {
      int n = 0;
      while (n < Rest && f(this[index + n])) n++;
      return Take(n);
    }

    public ByteReader TakeUntil(Func<byte, bool> f) {
      int n = 0;
      while (n < Rest && !f(this[index + n])) n++;
      if (n < Rest) n++;
      return Take(n);
    }

    public ByteReader Drop(int n) {
      if (n < 0 || Rest < n) throw new ArgumentOutOfRangeException();
      return new ByteReader(source, offset + index + n, Rest - n);
    }

    public ByteReader DropWhile(Func<byte, bool> f) {
      int n = 0;
      while (n < Rest && f(this[index + n])) n++;
      return Drop(n);
    }

    public ByteReader DropUntil(Func<byte, bool> f) {
      int n = 0;
      while (n < Rest && !f(this[index + n])) n++;
      if (n < Rest) n++;
      return Drop(n);
    }

    public void Skip(int n) {
      if (n > Rest) throw new ArgumentOutOfRangeException();
      index += n;
    }

    public short ShortOnBigEndian() {
      if (Rest < 2) throw new InvalidOperationException();
      short val = (short)((this[index] << 8) + this[index + 1]);
      index += 2;
      return val;
    }
    public short ShortOnLittleEndian() {
      if (Rest < 2) throw new InvalidOperationException();
      short val = (short)((this[index + 1] << 8) + this[index]);
      index += 2;
      return val;
    }

    public int IntOnBigEndian() {
      if (Rest < 4) throw new InvalidOperationException();
      int val = (this[index] << 24) + (this[index + 1] << 16) + (this[index + 2] << 8) + this[index + 3];
      index += 4;
      return val;
    }

    public int IntOnLittleEndian() {
      if (Rest < 4) throw new InvalidOperationException();
      int val = (this[index + 3] << 24) + (this[index + 2] << 16) + (this[index + 1] << 8) + this[index];
      index += 4;
      return val;
    }
  }
}
