﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using Microsoft.VisualBasic;
using San9.Resource.IO;

namespace San9.Resource.Message {
  public class TextMessage : Message {
    private static readonly Encoding encoding = Encoding.GetEncoding("Shift-JIS");

    private static readonly byte[] Suffix = new byte[] { 0x05, 0x05, 0x05 };

    private static readonly byte[] FormatPrefix = new byte[] { 0x02, 0x8C };
    private static readonly byte[] FormatSuffix = new byte[] { 0x62 };
    private static readonly byte[] FullKanaPrefix = new byte[] { 0x1B, 0x4B };
    private static readonly byte[] FullKanaSuffix = new byte[] { 0x1B, 0x48 };
    private static readonly byte[] HalfKanaPrefix = new byte[] { 0x1B, 0x6B };
    private static readonly byte[] HalfKanaSuffix = new byte[] { 0x1B, 0x48 };

    public readonly string Text;

    public TextMessage(string text) {
      this.Text = text;
    }

    public static TextMessage Parse(ByteReader source) {
      StringBuilder text = new StringBuilder();
      while (!check(source, Suffix)) {
        if (check(source, FormatPrefix)) {
          Number n = Number.Parse(source);
          if (n == null || !check(source, FormatSuffix)) throw new ArgumentException();
          text.Append("#{" + n.Text + "}");
        } else if (check(source, FullKanaPrefix)) {
          byte[] bytes = source.TakeWhile(b => isHalfKana(b)).ToArray();
          text.Append(Strings.StrConv(encoding.GetString(bytes), VbStrConv.Wide));
          source.Skip(bytes.Length);
          if (!check(source, FullKanaSuffix)) throw new ArgumentException();
        } else if (check(source, HalfKanaPrefix)) {
          byte[] bytes = source.TakeWhile(b => isHalfKana(b)).ToArray();
          text.Append(encoding.GetString(bytes));
          source.Skip(bytes.Length);
          if (!check(source, HalfKanaSuffix)) throw new ArgumentException();
        } else if (isFull(source.Current)) {
          int i = 0;
          byte[] bytes = source.TakeWhile(b => (++i % 2) == 0 || isFull(b)).ToArray();
          text.Append(encoding.GetString(bytes));
          source.Skip(bytes.Length);
        } else if (isHalfKana(source.Current)) {
          byte[] bytes = source.TakeWhile(b => isHalfKana(b)).ToArray();
          text.Append(Strings.StrConv(encoding.GetString(bytes), VbStrConv.Wide | VbStrConv.Hiragana));
          source.Skip(bytes.Length);
        } else if (Noun.Prefix == source.Current) {
          Noun n = Noun.Parse(source);
          if (n == null) throw new ArgumentException();
          text.Append("${" + n.Text + "}");
        } else if (0x0A == source.Current) {
          text.Append("\r\n");
          source.Skip(1);
        } else {
          text.Append(encoding.GetString(new byte[] { source.Current }));
          source.Skip(1);
        }
      }
      return new TextMessage(text.ToString());
    }

    private static bool isFull(byte b) {
      if (0x80 <= b && b <= 0xA0)
        return true;
      if (0xE0 <= b && b <= 0xFF)
        return true;
      return false;
    }

    private static bool isHalfKana(byte b) {
      if (0xA6 <= b && b <= 0xDF)
        return true;
      return false;
    }

    private static bool isHalfOther(byte b) {
      if (0x21 <= b && b <= 0x7E)
        return true;
      if (0xA1 <= b && b <= 0xA5)
        return true;
      if (b == 0x0D || b == 0x0A)
        return true;
      return false;
    }

    public override IEnumerable<byte> Source {
      get {
        Regex fullKana = new Regex(@"^([ア-ン]+)([^ア-ン]*)$");
        Regex halfKana = new Regex(@"^([ｱ-ﾝ]+)([^ｱ-ﾝ]*)$");
        Regex hiragana = new Regex(@"^([あ-ん]+)([^あ-ん]*)$");
        Regex format = new Regex(@"^#\{([0-9]+)\}(.*)$");
        Regex noun = new Regex(@"^\$\{(.+)\}(.*)$");
        string rest = Text.Replace("\r\n", "\n");
        while (rest != string.Empty) {
          if (fullKana.IsMatch(rest)) {
            Match m = fullKana.Match(rest);
            foreach (byte b in FullKanaPrefix)
              yield return b;
            foreach (byte b in encoding.GetBytes(Strings.StrConv(m.Groups[1].Value, VbStrConv.Narrow)))
              yield return b;
            foreach (byte b in FullKanaSuffix)
              yield return b;
            rest = m.Groups[2].Value;
          } else if (halfKana.IsMatch(rest)) {
            Match m = halfKana.Match(rest);
            foreach (byte b in HalfKanaPrefix)
              yield return b;
            foreach (byte b in encoding.GetBytes(m.Groups[1].Value))
              yield return b;
            foreach (byte b in HalfKanaSuffix)
              yield return b;
            rest = m.Groups[2].Value;
          } else if (hiragana.IsMatch(rest)) {
            Match m = hiragana.Match(rest);
            foreach (byte b in encoding.GetBytes(Strings.StrConv(m.Groups[1].Value, VbStrConv.Narrow | VbStrConv.Katakana)))
              yield return b;
            rest = m.Groups[2].Value;
          } else if (format.IsMatch(rest)) {
            Match m = format.Match(rest);
            foreach (byte b in FormatPrefix)
              yield return b;
            foreach (byte b in encoding.GetBytes(m.Groups[1].Value))
              yield return b;
            foreach (byte b in FormatSuffix)
              yield return b;
            rest = m.Groups[2].Value;
          } else if (noun.IsMatch(rest)) {
            Match m = noun.Match(rest);
            Noun n = new Noun(m.Groups[1].Value);
            foreach (byte b in n.Source)
              yield return b;
            rest = m.Groups[2].Value;
          } else {
            foreach (byte b in encoding.GetBytes(rest[0].ToString()))
              yield return b;
            rest = rest.Substring(1);
          }
        }
        foreach (byte b in Suffix)
          yield return b;
      }
    }
  }
}
